import ast
import json
import logging
import numbers
import threading
import uuid
from inspect import signature
from typing import Optional, Any, Callable

from pika import BlockingConnection, BasicProperties
from pika.adapters.blocking_connection import BlockingChannel
from pika.spec import Basic

from snacks.rabbit_config import RabbitConfig

__all__ = ('RabbitApp',)
log = logging.getLogger('snacks')


# TODO snacks-django, finds all serializers, uses those.
class RabbitApp:
    """A class to interface with RabbitMQ."""

    def __init__(self, rabbit_config: RabbitConfig) -> None:
        self.config = rabbit_config
        log.info('RabbitApp configured.')
        log.debug('host=%s', rabbit_config.host)
        log.debug('port=%s', rabbit_config.port)
        log.debug('exchange=%s', rabbit_config.exchange)
        log.debug('virtual_host=%s', rabbit_config.virtual_host)

    # TODO Make async?
    def publish(self, body: Any, routing_key: str, **kwargs) -> None:
        """Publish message to a rabbit queue with the given routing key.

        :param body: The message to publish.
        :param kwargs: Keyword args to pass to pika publish.
        :param routing_key: The routing key.
        """
        log.debug('Publishing key=%s msg=%s', routing_key, body)
        mq_conn = BlockingConnection(self.config.params)
        channel: BlockingChannel = mq_conn.channel()
        if 'to_json' in dir(body):
            body = body.to_json()
        elif isinstance(body, numbers.Number):
            body = str(body)
        channel.basic_publish(
            exchange=self.config.exchange,
            routing_key=routing_key,
            body=body,
            **kwargs
        )

    # TODO Make async?
    def publish_and_receive(
            self,
            body: Any,
            routing_key: str,
            serializer: Optional[Callable] = None,
            time_limit: int = 60,
            headers: Optional[dict[str, Any]] = None
    ) -> Any:
        """Publish message to a rabbit queue with the given routing key.

        :param body: The message to publish.
        :param routing_key: The routing key.
        :param serializer: Callable to serialize response bodies.
        :param time_limit: Number of seconds to wait for a response.
        :param headers: Headers to publish with.
        """
        log.debug('Publishing key=%s msg=%s', routing_key, body)
        mq_conn = BlockingConnection(self.config.params)
        channel: BlockingChannel = mq_conn.channel()
        response: Optional[bytes] = None
        corr_id = str(uuid.uuid4())

        def _on_response(
                _channel: BlockingChannel,
                _method: Basic.Deliver,
                props: BasicProperties,
                resp_body: bytes
        ) -> None:
            nonlocal response
            if props.correlation_id == corr_id:
                response = resp_body

        result = channel.queue_declare(queue='', exclusive=True)
        callback_queue = result.method.queue
        channel.basic_consume(
            queue=callback_queue,
            on_message_callback=_on_response,
            auto_ack=True
        )
        channel.basic_publish(
            exchange=self.config.exchange,
            routing_key=routing_key,
            properties=BasicProperties(
                reply_to=callback_queue,
                correlation_id=corr_id,
                headers=headers
            ),
            body=body
        )
        mq_conn.process_data_events(time_limit=time_limit)
        return serializer(response) if serializer else response

    def listener(self, queues: list[str], serializer=None) -> Callable:
        """Decorate a function to make it a rabbit consumer.

        The decorated function can have as parameters, the message body,
        the method, the properties, or no parameters.
        A new thread is made for each consumer.

        :param queues: The queues to consume from.
        :param serializer: Callable to serialize message bodies.
        :return: Function decorated to be a rabbit consumer.
        """
        mq_conn = BlockingConnection(self.config.params)
        channel = mq_conn.channel()

        def wrapper(fun: Callable) -> None:
            original_name = fun.__name__
            for queue in queues:
                fun.__name__ = f'listener_{queue}'
                _listen(channel, self.config.exchange, queue, fun, serializer)
                log.info('%s subscribed to %s', original_name, queue)
            thread = threading.Thread(target=_consume, args=(channel,))
            thread.daemon = True
            thread.start()

        return wrapper


def _listen(
        channel: BlockingChannel,
        exchange: str,
        queue: str,
        fun: Callable,
        serializer: Optional[Callable] = None
) -> None:
    def callback(
            ch: BlockingChannel,
            method: Basic.Deliver,
            properties: BasicProperties,
            body: bytes
    ) -> None:
        log.debug('%s received %s', queue, body)
        sig = signature(fun)
        kwargs: dict[str, Any] = {}
        for name, param in sig.parameters.items():
            if name == 'self':
                kwargs[name] = None
            elif param.annotation == BlockingChannel:
                kwargs[name] = ch
            elif param.annotation == method:
                kwargs[name] = method
            elif param.annotation == BasicProperties:
                kwargs[name] = properties
            elif param.annotation == str:
                kwargs[name] = body.decode('utf-8')
            elif param.annotation in {list, tuple, set, dict}:
                kwargs[name] = ast.literal_eval(body.decode('UTF-8'))
            elif param.annotation == bytes:
                kwargs[name] = body
            elif param.annotation in {int, float}:
                kwargs[name] = param.annotation(body)
            elif 'from_json' in dir(param.annotation):
                kwargs[name] = param.annotation.from_json(body)
            else:
                try:
                    kwargs[name] = json.loads(body)
                except json.decoder.JSONDecodeError:
                    kwargs[name] = body
        # noinspection PyBroadException
        try:
            resp = fun(**kwargs)
            if properties.reply_to:
                channel.basic_publish(
                    exchange=exchange,
                    routing_key=properties.reply_to,
                    properties=BasicProperties(
                        correlation_id=properties.correlation_id
                    ),
                    body=serializer(resp) if serializer else resp
                )
                channel.basic_ack(delivery_tag=method.delivery_tag)
        except Exception as e:
            log.exception(msg=f'{type(e).__name__}: {e}', exc_info=e)

    channel.basic_qos(prefetch_count=1)
    channel.basic_consume(
        queue=queue,
        auto_ack=True,
        on_message_callback=callback
    )


def _consume(channel: BlockingChannel) -> None:
    try:
        channel.start_consuming()
    except KeyboardInterrupt:
        channel.stop_consuming()
        channel.close()
